#
# Copyright 2020 Canonical Ltd.  All rights reserved.
#

import ipaddress
import os
import string
import random
from subprocess import check_call, check_output, STDOUT, CalledProcessError

import netifaces
from charmhelpers.core.hookenv import unit_public_ip, config, log, charm_name, close_port, open_port, open_ports, \
    close_ports, opened_ports

CURRENT_VERSION = "1.9"


def get_ip_for_interface(interface):
    """Returns the IP address associated to the interface"""
    addresses = netifaces.ifaddresses(interface)
    if (netifaces.AF_INET not in addresses or
            'addr' not in addresses[netifaces.AF_INET][0]):
        raise Exception('No IP associated to requested device')

    return addresses[netifaces.AF_INET][0]['addr']


def get_public_address():
    """
    Returns the juju unit public address.
    If 'public_interface' is configured, get its IP instead
    """
    public_address = unit_public_ip()
    public_interface = config('public_interface')
    if public_interface:
        public_address = get_ip_for_interface(public_interface)
        if not public_address:
            log('Could not obtain a valid IP for the configured public_interface. \
                Using the default one')
            public_address = unit_public_ip()
    return public_address


def remove_user_from_group(user, group):
    """Removes a system group from user groups"""
    log("Removing user {} from group {}".format(user, group), level="INFO")
    os.system('gpasswd -d {} {}'.format(user, group))


def remove_group(group):
    """Removes a group from the system"""
    log("Removing group {}".format(group), level="INFO")
    os.system('groupdel {}'.format(group))


def reload_systemd():
    """Reload systemd daemon"""
    check_call(['systemctl', 'daemon-reload'])


def generate_random_string(length):
    letters = string.ascii_letters + string.digits
    return ''.join(random.choice(letters) for _ in range(length))


def pin_package_version(name, version=CURRENT_VERSION):
    """Pin a deb package version to ensure controlled apt upgrades"""
    with open("/etc/apt/preferences.d/{}".format(name), "w+") as f:
        f.writelines([
            "Package: {}\n".format(name),
            "Pin: version {}.*\n".format(version),
            "Pin-Priority: 1000\n"])


def remove_package_pin(name):
    """Remove a pin introduced by calling pin_package_version"""
    pin_path = "/etc/apt/preferences.d/{}".format(name)
    if not os.path.exists(pin_path):
        return

    os.remove(pin_path)


def snap_set(param, value):
    """Sets a config parameter for this charm main snap"""
    try:
        cmd = ['snap', 'set', charm_name(), '{}={}'.format(param, value)]
        check_output(cmd, stderr=STDOUT)
    except CalledProcessError as ex:
        log('Could not set snap config param: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def close_old_and_open_new_ports(port_parameter, protocol='TCP'):
    """Close the previous port or port range and opens the new provided one"""
    previous_range = config().previous(port_parameter)
    new_range = config().get(port_parameter)

    if not new_range:
        return

    new_start = extract_range_start(new_range)
    new_end = extract_range_end(new_range)
    if new_start > new_end:
        raise Exception('Invalid interval {}'.format(new_range))

    if previous_range is not None and previous_range != new_range:
        prev_start = extract_range_start(previous_range)
        prev_end = extract_range_end(previous_range)
        log('Closing port: {}/{}'.format(previous_range, protocol), level='INFO')
        close_port_range(prev_start, prev_end, protocol)

    log('Opening port: {}/{}'.format(new_range, protocol), level='INFO')
    open_port_range(new_start, new_end, protocol=protocol)


def close_port_range(start, end, protocol='TCP'):
    if start == end:
        close_port(start, protocol=protocol)
        return
    close_ports(start, end, protocol=protocol)


def open_port_range(start, end, protocol='TCP'):
    if start == end:
        open_port(start, protocol=protocol)
        return
    open_ports(start, end, protocol=protocol)


def extract_range_start(r):
    return extract_range_limit(r, False)


def extract_range_end(r):
    return extract_range_limit(r, True)


def extract_range_limit(r, is_end_limit):
    if type(r) is int:
        return r
    limit = r
    dash_pos = r.find('-')
    if dash_pos != -1:
        if is_end_limit:
            limit = r[dash_pos+1:]
        else:
            limit = r[:dash_pos]
    return int(limit)

def is_ip_address(address):
    try:
        ipaddress.ip_address(address)
        return True
    except:
        return False