#
# Anbox - The Android in a Box runtime environment
# Copyright 2019 Canonical Ltd.  All rights reserved.
#

import os
import shutil
import subprocess
from subprocess import check_output, check_call, CalledProcessError, STDOUT

from charmhelpers import fetch
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.contrib.network.ip import get_host_ip
from charmhelpers.core import templating
from charmhelpers.core.hookenv import (unit_private_ip, config, log,
                                       status_set, relation_ids, unit_get,
                                       application_version_set)
from charmhelpers.core.host import service_restart
from charms.layer.anbox import (pin_package_version, remove_package_pin)
from charms.layer.pki import (create_cert_files_if_needed, read_cert, fingerprint)
from charms.reactive import hook, when, when_any, when_not
from charms.reactive.flags import clear_flag, is_flag_set, set_flag
from charms.reactive.relations import endpoint_from_flag
from charms.layer.anbox import close_old_and_open_new_ports

SERVICE = "snap.aar.daemon.service"

SNAP_COMMON_PATH = "/var/snap/aar/common"

AAR_CONFIG_PATH = "{}/conf/main.yaml".format(SNAP_COMMON_PATH)

AAR_CERT_BASE_PATH = "{}/certs".format(SNAP_COMMON_PATH)

CLIENT_CERT_PATH = "{}/clients/".format(AAR_CERT_BASE_PATH)
PUBLISHERS_CERT_PATH = "{}/publishers/".format(AAR_CERT_BASE_PATH)
AAR_SERVER_CERT_PATH = "{}/server.crt".format(AAR_CERT_BASE_PATH)
AAR_SERVER_KEY_PATH = "{}/server.key".format(AAR_CERT_BASE_PATH)

DEB_PACKAGE_NAME = 'anbox-cloud-aar'


@when('ua.ready')
@when_not('aar.installed')
def install():
    pin_package_version(DEB_PACKAGE_NAME)
    fetch.apt_install([DEB_PACKAGE_NAME], fatal=True)
    set_charm_version()

    check_call(['snap', 'connect', 'aar:home', ':home'])
    check_call(['snap', 'connect', 'aar:network', ':network'])

    os.makedirs(CLIENT_CERT_PATH, 0o700, exist_ok=True)
    os.makedirs(PUBLISHERS_CERT_PATH, 0o700, exist_ok=True)
    create_cert_files_if_needed(AAR_SERVER_CERT_PATH, AAR_SERVER_KEY_PATH)
    status_set('active', '')
    set_flag('aar.installed')

    configure()
    set_flag('aar.ready')


@hook('stop')
def stop():
    """Remove the charm main snap when charm is removed"""

    if not is_flag_set('aar.installed'):
        return

    fetch.apt_purge([DEB_PACKAGE_NAME], fatal=True)
    remove_package_pin(DEB_PACKAGE_NAME)


@hook('upgrade-charm')
def upgrade_charm():
    if not is_flag_set('aar.installed'):
        return

    # This will rewrite out pin to a potential newer version of the
    # debian package we're going to upgrade to
    pin_package_version(DEB_PACKAGE_NAME)

    fetch.apt_update(fatal=True)
    fetch.apt_install([DEB_PACKAGE_NAME], fatal=True)
    set_charm_version()

    configure()

    # If we have an NRPE relation, force run of that as well
    if relation_ids('nrpe-exteral-master'):
        clear_flag('nagios.configured')


def set_charm_version():
    try:
        info = subprocess.run(['snap', 'info', 'aar'],
                              stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as err:
        log('Failed to get snap info: {}'.format(err), level='ERROR')
    else:
        stdout = info.stdout.decode('utf8').splitlines()
        version = [line.split()[1] for line in stdout if 'installed' in line]
        if version:
            application_version_set(version[0])


@when_any('config.changed')
def configure():
    if not is_flag_set('aar.installed'):
        return

    status_set('maintenance', 'Update configuration')
    port = config('port')
    listen_address = unit_private_ip()

    templating.render(
        source='config.yaml',
        target=AAR_CONFIG_PATH,
        context={
            'listen_address': "{}:{}".format(listen_address, port),
            'storage_config': config("storage_config")
        },
        perms=0o600)

    close_old_and_open_new_ports('port')

    service_restart(SERVICE)

    clients = endpoint_from_flag('endpoint.client.joined')
    if clients is not None:
        endpoint_aar_joined(clients)

    publishers = endpoint_from_flag('endpoint.publisher.joined')
    if publishers is not None:
        endpoint_aar_joined(publishers)

    status_set('active', '')


@when('endpoint.publisher.joined', 'aar.ready')
def endpoint_arr_push_joined():
    aar = endpoint_from_flag('endpoint.publisher.joined')
    endpoint_aar_joined(aar)


@when('endpoint.client.joined', 'aar.ready')
def endpoint_arr_pull_joined():
    aar = endpoint_from_flag('endpoint.client.joined')
    endpoint_aar_joined(aar)


def endpoint_aar_joined(aar):
    cert = read_cert(AAR_SERVER_CERT_PATH)
    if cert == '':
        status_set('blocked', 'No registry certificate')
        return

    listen_address = unit_private_ip()
    location = config('location')
    if len(location) > 0:
        listen_address = location

    aar.set_registry_information(
        certificate=cert,
        fingerprint=fingerprint(cert),
        ip=listen_address,
        port=config('port')
    )


@when_any('endpoint.publisher.changed',
          'endpoint.publisher.departed')
@when('aar.ready')
def endpoint_arr_push_changed():
    aar = endpoint_from_flag('endpoint.publisher.changed')
    if aar is not None:
        endpoint_aar_changed(aar, 'publisher')

    if is_flag_set('endpoint.publisher.changed'):
        clear_flag('endpoint.publisher.changed')

    if is_flag_set('endpoint.publisher.departed'):
        clear_flag('endpoint.publisher.deparated')


@when_any('endpoint.client.changed',
          'endpoint.client.departed')
@when('aar.ready')
def endpoint_arr_pull_changed():
    aar = endpoint_from_flag('endpoint.client.changed')
    if aar is not None:
        endpoint_aar_changed(aar, 'client')

    if is_flag_set('endpoint.client.changed'):
        clear_flag('endpoint.client.changed')

    if is_flag_set('endpoint.client.departed'):
        clear_flag('endpoint.client.deparated')


def endpoint_aar_changed(aar, mode):
    remove_all_certificates()
    ams_clients = aar.ams_clients()
    if ams_clients is None or len(ams_clients) == 0:
        return

    for client in ams_clients:
        register_ams_client(client['certificate'], mode)
    service_restart(SERVICE)


def remove_all_certificates():
    """
    All client certificates are kept by juju in the relation until the hook
    fails, so every time a new client is registered, the previous certificates
    will be added again. To avoid duplicates and properly take into account
    units that departed, we remove all certificates and add the ones that
    are still active again
    """
    shutil.rmtree(CLIENT_CERT_PATH)
    shutil.rmtree(PUBLISHERS_CERT_PATH)
    os.makedirs(CLIENT_CERT_PATH, 0o700, exist_ok=True)
    os.makedirs(PUBLISHERS_CERT_PATH, 0o700, exist_ok=True)


def register_ams_client(client_certificate, mode):
    extra_args = []
    if mode == 'publisher':
        extra_args = ["--publisher"]

    status_set('maintenance', 'Configuring new AMS client')
    try:
        os.remove('/root/client.crt')
    except OSError:
        pass

    # TODO: feed certificate directly to AAR via standard input
    # rather than by creating a file
    f = open('/root/client.crt', 'x')
    f.write(client_certificate)
    f.close()

    try:
        check_output(['/snap/bin/aar', 'trust', 'add', '/root/client.crt',
                      *extra_args], stderr=STDOUT)
    except CalledProcessError as ex:
        status_set('blocked', 'Failed to register client certificate')
        raise ex

    os.remove('/root/client.crt')
    status_set('active', '')


@when_any('config.changed.nagios_context',
          'config.changed.nagios_servicegroups')
def nagios_config_changed():
    if not is_flag_set('endpoint.nrpe-external-master.joined'):
        return
    update_nrpe_config()


@when('nrpe-external-master.available')
@when_not('nagios.configured')
def update_nrpe_config():
    hostname = nrpe.get_nagios_hostname()
    address = get_host_ip(unit_get('private-address'))
    port = config('port')
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    cmd = [
        '/usr/lib/nagios/plugins/check_http',
        '--ssl', '--hostname', "{}:{}".format(address, port),
        '--url', '/1.0/status', '--expect', '200',
        '--warning', '5', '--critical', '10'
    ]
    nrpe_setup.add_check(
        shortname='aar_https',
        description='Check AAR https endpoint',
        check_cmd=" ".join(cmd),
    )
    nrpe_setup.write()
    set_flag('nagios.configured')
